(* $Id: walk_read.ml,v 1.5 2012/06/18 00:01:15 weis Exp $ *)

let get_keywords, add_keyword, _reset_keyword_list =
  let keywords = ref [] in
  (fun () -> !keywords),
  (fun k -> keywords := k :: ! keywords),
  (fun () -> keywords := [])
;;

let get_main_parser_name, _set_main_parser_name =
  let main_parser_name = ref "parser" in
  (fun () -> !main_parser_name),
  (fun pname -> main_parser_name := pname)
;;

(* Generation of readers. *)
module Read : Action_spec.Make =
  functor (Io : Walk_spec.Out) ->

  struct

    let ppf = Io.buffer
    ;;

    let print_seq () =
      Format.fprintf ppf ";@ "
    ;;

    let meta_read_char c = Format.printf "Scanf.scanf ib \"%c\"" c
    ;;

    let meta_read_string s =
      Format.fprintf ppf "@[<1>Scanf.bscanf ib@ \"%s\"@]" (String.escaped s)
    ;;

    let add_sum_tag_rule _s = ();;

    let sum_tag lab =
        add_keyword lab;
        add_sum_tag_rule lab
    ;;

    let begin_sum_inj _lab = ();;
    let end_sum_inj _s = ();;

    let begin_tuple () = meta_read_char '('; print_seq ();;
      let begin_tuple_elem () = ();;
      let end_tuple_elem () = ();;
      let separator_tuple () = print_seq (); meta_read_char ','; print_seq ();;
    let end_tuple () = print_seq (); meta_read_char ')';;

    let begin_record_expr () = meta_read_char '{'; print_seq ();;
    let end_record_expr () = meta_read_char '}';;

    let begin_record_field lab =
      meta_read_string " "; print_seq ();
      meta_read_string lab; print_seq ();
      meta_read_string " = "; print_seq ();;
    let end_record_field _lab =
      print_seq (); meta_read_char ';'; print_seq ();;

    let begin_application () =
      Format.fprintf ppf "@[<1>"
    ;;
      let separator_argument () = Format.fprintf ppf "@ "
      ;;
      let begin_argument () = ()
      ;;
      let end_argument () = ()
      ;;
    let end_application () =
      Format.fprintf ppf "@]"
    ;;

    let external_fun = function
      | "read_unit" -> Format.fprintf ppf "Lib_read.read_unit"
      | "read_bool" -> Format.fprintf ppf "Lib_read.read_bool"
      | "read_string" -> Format.fprintf ppf "Lib_read.read_quoted_string"
      | "read_int" -> Format.fprintf ppf "Lib_read.read_quoted_int"
      | "read_float" -> Format.fprintf ppf "Lib_read.read_quoted_float"
      | "read_char" -> Format.fprintf ppf "Lib_read.read_quoted_char"
      | "read_list" -> Format.fprintf ppf "Lib_read.read_List"
      | "read_array" -> Format.fprintf ppf "Lib_read.read_Array"
      | "read_option" -> Format.fprintf ppf "Lib_read.read_Option"
      | "read_fun" -> Format.fprintf ppf "Lib_read.read_fun"
      | "read_poly" -> Format.fprintf ppf "Lib_read.read_poly"
      | s -> Format.fprintf ppf "%s" s
    ;;

    let begin_paren () =
      meta_read_string "("; print_seq ()
    ;;

    let end_paren () =
      meta_read_string ")"; print_seq ()
    ;;

  end
;;

(* Generator's concern: generates pattern matchings and recursive calls to
   the functions that walk the code. *)

module Gen : Walk_spec.Make =
  functor (Io : Walk_spec.Out) ->

  struct

    let ppf = Io.buffer;;

    let comment s =
      (* Sometimes the comments from the .mli file are meaningful for
         Ocamldoc in the .mli source file, but becomes irrelevant in the .ml
         target file! *)
      if Config.get_keep_comments () &&
      (* Comments have already been printer by the data type printer. *)
         not (Config.get_keep_type_definitions ())
      then Format.fprintf ppf "%s@." (Og_types_pprint.comment s)
    ;;
    let directive s = Format.fprintf ppf "open %s;;@.@." s;;

    let begin_file () =
      Format.fprintf ppf
        "(* Beginning of %s generated code *)@ @ "
        (Filename.basename Sys.executable_name);
      if not (Config.get_keep_type_definitions ()) then
         directive (Config.get_input_module_name ())
    ;;

    let quoted_escaped_string s = "\"" ^ String.escaped s ^ "\"";;

    let string_list ppf = function
    | [] -> ()
    | s :: sl ->
      Format.fprintf ppf "%s" (quoted_escaped_string s);
      List.iter
        (fun s -> Format.fprintf ppf ";@ %s" (quoted_escaped_string s)) sl
    ;;

    let end_file () =
      let systeme_kwds =
        [ "("; ")"; "{"; "}"; "["; "]"; "[|"; "|]"; ","; ";";
          "true"; "false" ] in
      Format.fprintf ppf
        "let lexer =@. make_lexer @[[%a]@];;@.@."
        (string_list) (systeme_kwds @ (get_keywords ()));
      let parser_name = get_main_parser_name () in
      Format.fprintf ppf
         "\
         let lexem_stream ic =@.  \
           let str = Stream.of_channel ic in@.  \
           Stream.from (fun x -> Some (lexer str))@.\
         ;;@.@.\
         let parse_channel ic =@.  \
           %s (lexem_stream ic);;@.@.\
         let parse_string s =@.  \
           %s (lexer (Stream.of_string s))@."
        parser_name parser_name;

      Format.fprintf ppf
        "(* End of %s generated code *)@."
        Sys.executable_name
    ;;

    let variable s = Format.fprintf ppf "%s" s;;

    let begin_paren () = Format.fprintf ppf "@[(";;
    let end_paren () = Format.fprintf ppf "@,)@]";;

    let begin_application () = Format.fprintf ppf "@[<1>";;
      let separator_argument () = Format.fprintf ppf "@ ";;
      let begin_argument () = Format.fprintf ppf "(";;
      let end_argument () = Format.fprintf ppf ")";;
    let end_application () = Format.fprintf ppf "@]";;

    let begin_tuple () = Format.fprintf ppf "@[<1>(";;
      let separator_tuple () = Format.fprintf ppf ",@ ";;
    let end_tuple () = Format.fprintf ppf ")@]";;

    let begin_record_pattern () = Format.fprintf ppf "@[{@[";;
      let begin_record_field f = Format.fprintf ppf "@[<1>%s =@ " f;;
      let end_record_field () = Format.fprintf ppf "@]";;
      let separator_record_field () = Format.fprintf ppf ";@ ";;
    let end_record_pattern () = Format.fprintf ppf "@]@,}@] ->@,";;

    let begin_def rec_flag =
      Format.fprintf ppf "@[<v 0>le@[t %s"
        (if rec_flag = Walker_ast.Rec then "rec " else "")
    ;;
      let begin_def_elem ((_ty_name, walker_name), ty_arguments) =
        variable walker_name;
        List.iter (function arg -> Format.fprintf ppf " %s" arg) ty_arguments;
        Format.fprintf ppf " =@ "
      ;;
      let end_def_elem () = Format.fprintf ppf "@]";;
      let separator_def_elem () = Format.fprintf ppf "@,@,an@[d ";;
    let end_def () = Format.fprintf ppf "@,;;@]@]@.@.";;

    let begin_function () = Format.fprintf ppf "@[<1>(function@ ";;
      let begin_clause_pattern () = Format.fprintf ppf "@[";;
      let end_clause_pattern () = Format.fprintf ppf "@] ->";;
      let begin_clause_expr () = Format.fprintf ppf "@;<1 1>@[";;
      let end_clause_expr () =  Format.fprintf ppf "@]";;
    let end_function () = Format.fprintf ppf ")@]";;

    let sum_tag s = add_keyword s; Format.fprintf ppf "%s" s;;

    let begin_sum_inj s = add_keyword s; Format.fprintf ppf "%s " s;;
    let end_sum_inj _s = ();;

    let begin_sum_clauses () = Format.fprintf ppf "function@;<1 -1>@[";;
      let begin_sum_clause () = Format.fprintf ppf "@;<1 1>| @[";;
      let end_sum_clause () = Format.fprintf ppf "@]";;
    let end_sum_clauses () = Format.fprintf ppf "@]";;

    let begin_alias_clause () = Format.fprintf ppf "@[<1>";;
      let begin_aliased_clause () = Format.fprintf ppf "@[<1>";;
      let end_aliased_clause () = Format.fprintf ppf "@]";;
    let end_alias_clause () = Format.fprintf ppf "@]";;

    let begin_record_clause () = Format.fprintf ppf "function@,@[<1>";;
    let end_record_clause () = Format.fprintf ppf "@]";;

  end
;;

(* Generates a value reading program. *)
module Reader =
  functor (Io : Walk_spec.Out) ->
  Walker.Make (Gen) (Read) (Io)
;;
